import json

from oauthlib.oauth2 import MobileApplicationClient, BackendApplicationClient, WebApplicationClient
from requests.auth import HTTPBasicAuth
from requests_oauthlib import OAuth2Session
from future.builtins import bytes

from mendeley.session import MendeleySession


def handle_text_response(rsp):
    if rsp.headers['content-type'] == 'text/plain':
        rsp._content = bytes(json.dumps({'error': 'invalid_client', 'error_description': rsp.text}), rsp.encoding)
        rsp.headers['content-type'] = 'application/json'

    return rsp


class MendeleyClientCredentialsAuthenticator(object):
    def __init__(self, mendeley):
        self.mendeley = mendeley

        self.oauth = OAuth2Session(
            client=BackendApplicationClient(mendeley.client_id),
            scope=['all'])
        self.oauth.compliance_hook['access_token_response'] = [handle_text_response]

    def authenticate(self):
        token_url = self.mendeley.host + '/oauth/token'
        auth = HTTPBasicAuth(self.mendeley.client_id, self.mendeley.client_secret)

        token = self.oauth.fetch_token(token_url, auth=auth, scope=['all'])
        return MendeleySession(self.mendeley, token)


class MendeleyLoginAuthenticator:
    def __init__(self, mendeley, client, state):
        self.mendeley = mendeley
        self.state = state

        self.oauth = OAuth2Session(
            client=client,
            redirect_uri=mendeley.redirect_uri,
            scope=['all'],
            state=state)
        self.oauth.compliance_hook['access_token_response'] = [handle_text_response]

    def get_login_url(self):
        base_url = self.mendeley.host + '/oauth/authorize'
        (login_url, state) = self.oauth.authorization_url(base_url)
        return login_url


class MendeleyAuthorizationCodeAuthenticator(MendeleyLoginAuthenticator):
    def __init__(self, mendeley, state):
        client = WebApplicationClient(mendeley.client_id)
        MendeleyLoginAuthenticator.__init__(self, mendeley, client, state)

    def authenticate(self, redirect_url):
        token_url = self.mendeley.host + '/oauth/token'
        auth = HTTPBasicAuth(self.mendeley.client_id, self.mendeley.client_secret)

        token = self.oauth.fetch_token(token_url, authorization_response=redirect_url, auth=auth, scope=['all'])
        return MendeleySession(self.mendeley, token)


class MendeleyImplicitGrantAuthenticator(MendeleyLoginAuthenticator):
    def __init__(self, mendeley, state):
        client = MobileApplicationClient(mendeley.client_id)
        MendeleyLoginAuthenticator.__init__(self, mendeley, client, state)

    def authenticate(self, redirect_url):
        token = self.oauth.token_from_fragment(redirect_url)
        return MendeleySession(self.mendeley, token)



from mendeley.models.documents import TrashAllDocument, TrashBibDocument, TrashClientDocument, TrashTagsDocument, \
    TrashDocument
from mendeley.resources.base_documents import DocumentsBase


class Trash(DocumentsBase):
    """
    Top-level resource for accessing trashed documents.  These can be:

    - trashed documents for the logged-in user, if retrieved from a
      :func:`MendeleySession <mendeley.session.MendeleySession.trash>`.
    - trashed documents in a :func:`Group <mendeley.models.groups.Group.trash>`.
    """
    _url = '/trash'

    def __init__(self, session, group_id):
        super(Trash, self).__init__(session, group_id)

    def get(self, id, view=None):
        """
        Retrieves a trashed document by ID.

        :param id: the ID of the document to get.
        :param view: the view to get.  One of 'bib', 'client', 'tags', 'all'.
        :return: a :class:`TrashDocument <mendeley.models.documents.TrashDocument>`.
        """
        return super(Trash, self).get(id, view)

    def list(self, page_size=None, view=None, sort=None, order=None, modified_since=None, deleted_since=None):
        """
        Retrieves trashed documents, as a paginated collection.

        :param page_size: the number of documents to return on each page.  Defaults to 20.
        :param view: the view to get.  One of 'bib', 'client', 'tags', 'all'.
        :param sort: if specified, sorts documents by the specified field.  One of 'created', 'last_modified', 'title'.
        :param order: if specified in conjunction with 'sort', specifies the sort order.  One of 'asc', 'desc'.
        :param modified_since: if specified, only returns files modified after this timestamp.
        :param deleted_since: if specified, only returns the IDs of documents deleted after this timestamp.
        :return: a :class:`Page <mendeley.pagination.Page>` of
                 :class:`TrashDocuments <mendeley.models.documents.TrashDocument>`.
        """
        return super(Trash, self).list(page_size, view, sort, order, modified_since, deleted_since)

    def iter(self, page_size=None, view=None, sort=None, order=None, modified_since=None, deleted_since=None):
        """
        Retrieves trashed documents, as an iterator.

        :param page_size: the number of documents to retrieve at a time.  Defaults to 20.
        :param view: the view to get.  One of 'bib', 'client', 'tags', 'all'.
        :param sort: if specified, sorts documents by the specified field.  One of 'created', 'last_modified', 'title'.
        :param order: if specified in conjunction with 'sort', specifies the sort order.  One of 'asc', 'desc'.
        :param modified_since: if specified, only returns files modified after this timestamp.
        :param deleted_since: if specified, only returns the IDs of documents deleted after this timestamp.
        :return: an iterator of :class:`TrashDocuments <mendeley.models.documents.TrashDocument>`.
        """
        return super(Trash, self).iter(page_size, view, sort, order, modified_since, deleted_since)

    @staticmethod
    def _view_type(view):
        return {
            'all': TrashAllDocument,
            'bib': TrashBibDocument,
            'client': TrashClientDocument,
            'tags': TrashTagsDocument,
            'core': TrashDocument,
        }.get(view, TrashDocument)



class MendeleyException(Exception):
    pass


class MendeleyApiException(MendeleyException):
    def __init__(self, rsp):
        self.rsp = rsp

    def __str__(self):
        return 'The Mendeley API returned an error (status: %s, message: %s)' % (self.status, self.message)

    @property
    def status(self):
        return self.rsp.status_code

    @property
    def message(self):
        try:
            return self.rsp.json()['message']
        except ValueError:
            return self.rsp.text



from mendeley.models.common import Person
from mendeley.response import SessionResponseObject


class BaseDocument(SessionResponseObject):
    content_type = 'application/vnd.mendeley-document.1+json'

    @property
    def authors(self):
        """
        a list of :class:`Persons <mendeley.models.common.Person>`.
        """
        if 'authors' in self.json:
            return [Person(p) for p in self.json['authors']]
        else:
            return None

    @classmethod
    def fields(cls):
        return ['id', 'title', 'type', 'source', 'year', 'identifiers', 'keywords', 'abstract']


class BaseClientView(SessionResponseObject):
    @classmethod
    def fields(cls):
        return ['file_attached']


class BaseBibView(SessionResponseObject):
    @property
    def editors(self):
        """
        a list of :class:`Persons <mendeley.models.common.Person>`.
        """
        if 'editors' in self.json:
            return [Person(p) for p in self.json['editors']]
        else:
            return None

    @classmethod
    def fields(cls):
        return ['pages', 'volume', 'issue', 'websites', 'month', 'publisher', 'day', 'city', 'edition', 'institution',
                'series', 'chapter', 'revision']



from mendeley.models.groups import Group, GroupMember

from mendeley.resources.base import ListResource, GetByIdResource


class Groups(GetByIdResource, ListResource):
    """
    Top-level resource for accessing groups.
    """
    _url = '/groups'

    def __init__(self, session):
        self.session = session

    def get(self, id):
        """
        Retrieves a group by ID.

        :param id: the ID of the group to get.
        :return: a :class:`Group <mendeley.models.groups.Group>`.
        """
        return super(Groups, self).get(id)

    def list(self, page_size=None):
        """
        Retrieves groups that the logged-in user is a member of, as a paginated collection.

        :param page_size: the number of groups to return on each page.  Defaults to 20.
        :return: a :class:`Page <mendeley.pagination.Page>` of :class:`Groups <mendeley.models.groups.Group>`.
        """
        return super(Groups, self).list(page_size)

    def iter(self, page_size=None):
        """
        Retrieves groups that the logged-in user is a member of, as an iterator.

        :param page_size: the number of groups to retrieve at a time.  Defaults to 20.
        :return: an iterator of :class:`Groups <mendeley.models.groups.Group>`.
        """
        return super(Groups, self).iter(page_size)

    @property
    def _session(self):
        return self.session

    def _obj_type(self, **kwargs):
        return Group


class GroupMembers(ListResource):
    """
    Resource for accessing members of a group.
    """
    def __init__(self, session, id):
        self.session = session
        self.id = id

    def list(self, page_size=None):
        """
        Retrieves members of the group, as a paginated collection.

        :param page_size: the number of members to return on each page.  Defaults to 20.
        :return: a :class:`Page <mendeley.pagination.Page>` of
                 :class:`GroupMembers <mendeley.models.groups.GroupMember>`.
        """
        return super(GroupMembers, self).list(page_size)

    def iter(self, page_size=None):
        """
        Retrieves members of the group, as an iterator.

        :param page_size: the number of members to retrieve at a time.  Defaults to 20.
        :return: an iterator of :class:`GroupMembers <mendeley.models.groups.GroupMember>`.
        """
        return super(GroupMembers, self).iter(page_size)

    @property
    def _session(self):
        return self.session

    def _obj_type(self, **kwargs):
        return GroupMember

    @property
    def _url(self):
        return '/groups/%s/members' % self.id


from mendeley.models.base_documents import BaseDocument, BaseClientView, BaseBibView
from mendeley.response import LazyResponseObject


class CatalogDocument(BaseDocument):
    """
    Base class for catalog documents.

    .. attribute:: id
    .. attribute:: title
    .. attribute:: type
    .. attribute:: source
    .. attribute:: year
    .. attribute:: identifiers
    .. attribute:: keywords
    .. attribute:: abstract
    .. attribute:: link
    """
    @property
    def files(self):
        """
        a :class:`Files <mendeley.resources.files.Files>` resource, from which
        :class:`Files <mendeley.models.files.File>` can be retrieved.
        """
        return self.session.catalog_files(catalog_id=self.id)

    @classmethod
    def fields(cls):
        return super(CatalogDocument, cls).fields() + ['link']


class CatalogBibView(BaseBibView):
    """
    Additional fields returned when getting a :class:`CatalogDocument <mendeley.models.catalog.CatalogDocument>` with
    view='bib' or 'all'.

    .. attribute:: pages
    .. attribute:: volume
    .. attribute:: issue
    .. attribute:: websites
    .. attribute:: month
    .. attribute:: publisher
    .. attribute:: day
    .. attribute:: city
    .. attribute:: edition
    .. attribute:: institution
    .. attribute:: series
    .. attribute:: chapter
    .. attribute:: revision
    """
    pass


class CatalogClientView(BaseClientView):
    """
    Additional fields returned when getting a :class:`CatalogDocument <mendeley.models.catalog.CatalogDocument>` with
    view='client' or 'all'.

    .. attribute:: file_attached
    """
    pass


class CatalogStatsView(object):
    """
    Additional fields returned when getting a :class:`CatalogDocument <mendeley.models.catalog.CatalogDocument>` with
    view='stats' or 'all'.

    .. attribute:: reader_count
    .. attribute:: reader_count_by_academic_status
    .. attribute:: reader_count_by_subdiscipline
    .. attribute:: reader_count_by_country
    """
    @classmethod
    def fields(cls):
        return ['reader_count', 'reader_count_by_academic_status', 'reader_count_by_subdiscipline',
                'reader_count_by_country']


class CatalogBibDocument(CatalogBibView, CatalogDocument):
    @classmethod
    def fields(cls):
        return CatalogDocument.fields() + CatalogBibView.fields()


class CatalogClientDocument(CatalogClientView, CatalogDocument):
    @classmethod
    def fields(cls):
        return CatalogDocument.fields() + CatalogClientView.fields()


class CatalogStatsDocument(CatalogStatsView, CatalogDocument):
    @classmethod
    def fields(cls):
        return CatalogDocument.fields() + CatalogStatsView.fields()


class CatalogAllDocument(CatalogBibView, CatalogClientView, CatalogStatsView, CatalogDocument):
    @classmethod
    def fields(cls):
        return CatalogDocument.fields() + \
            CatalogBibView.fields() + \
            CatalogClientView.fields() + \
            CatalogStatsView.fields()


class LookupResponse(LazyResponseObject):
    def __init__(self, session, json, view, obj_type):
        super(LookupResponse, self).__init__(session, json['catalog_id'], obj_type, lambda: self._load())
        self.score = json['score']
        self.view = view

    def _load(self):
        return self.session.catalog.get(self.id, view=self.view)



import os
import re

from mendeley.response import SessionResponseObject


class File(SessionResponseObject):
    """
    A file attached to a document.

    .. attribute:: id
    .. attribute:: size
    .. attribute:: file_name
    .. attribute:: mime_type
    .. attribute:: filehash
    .. attribute:: download_url
    """
    content_type = 'application/vnd.mendeley-file.1+json'
    filename_regex = re.compile('filename="(\S+)"')

    @property
    def download_url(self):
        """
        the URL at which the file can be downloaded.  This is only valid for a short time, so should not be cached.
        """
        file_url = '/files/%s' % self.id
        rsp = self.session.get(file_url, allow_redirects=False)
        return rsp.headers['location']

    def document(self, view=None):
        """
        :param view: document view to return.
        :return: a :class:`UserDocument <mendeley.models.documents.UserDocument>` or
                 :class:`CatalogDocument <mendeley.models.catalog.CatalogDocument>`, depending on which the document is
                 attached to.
        """
        if 'document_id' in self.json:
            return self.session.documents.get_lazy(self.json['document_id'], view=view)
        elif 'catalog_id' in self.json:
            return self.session.catalog.get_lazy(self.json['catalog_id'], view=view)
        else:
            return None

    def download(self, directory):
        """
        Downloads the file.

        :param directory: the directory to download the file to.  This must exist.
        :return: the path to the downloaded file.
        """
        rsp = self.session.get('/files/%s' % self.id, stream=True)
        filename = self.filename_regex.search(rsp.headers['content-disposition']).group(1)
        path = os.path.join(directory, filename)

        with open(path, 'wb') as f:
            for block in rsp.iter_content(1024):
                if not block:
                    break

                f.write(block)

        return path

    def delete(self):
        """
        Deletes the file.
        """
        self.session.delete('/files/%s' % self.id)

    @classmethod
    def fields(cls):
        return ['id', 'size', 'file_name', 'mime_type', 'filehash']



from future.moves.urllib.parse import urlsplit, parse_qs, urlencode, urlunsplit
from future.utils import iteritems

from mendeley.pagination import Page
from mendeley.response import LazyResponseObject


class BaseResource(object):
    @property
    def _session(self):
        raise NotImplementedError

    @property
    def _url(self):
        raise NotImplementedError

    def _obj_type(self, **kwargs):
        raise NotImplementedError


class GetByIdResource(BaseResource):
    def get(self, id, **kwargs):
        url = add_query_params('%s/%s' % (self._url, id), kwargs)
        obj_type = self._obj_type(**kwargs)

        rsp = self._session.get(url, headers={'Accept': obj_type.content_type})

        return obj_type(self._session, rsp.json())

    def get_lazy(self, id, **kwargs):
        return LazyResponseObject(self._session, id, self._obj_type(**kwargs), lambda: self.get(id, **kwargs))


class ListResource(BaseResource):
    def list(self, page_size=None, **kwargs):
        return self._list(page_size, **kwargs)

    def iter(self, page_size=None, **kwargs):
        page = self._list(page_size, **kwargs)

        while page:
            for item in page.items:
                yield item

            page = page.next_page

    def _list(self, page_size, **kwargs):
        obj_type = self._obj_type(**kwargs)
        kwargs['limit'] = page_size
        url = add_query_params(self._url, kwargs)
        rsp = self._session.get(url, headers={'Accept': obj_type.content_type})
        return Page(self._session, rsp, obj_type)


def add_query_params(url, params):
    scheme, netloc, path, query_string, fragment = urlsplit(url)
    query_params = parse_qs(query_string)

    for name, value in iteritems(params):
        if value:
            query_params[name] = [value]

    new_query_string = urlencode(query_params, doseq=True)

    return urlunsplit((scheme, netloc, path, new_query_string, fragment))



from mendeley.exception import MendeleyException
from mendeley.models.catalog import *
from mendeley.resources.base import add_query_params, ListResource, GetByIdResource


class Catalog(GetByIdResource):
    """
    Top-level resource for accessing catalog documents.
    """
    _url = '/catalog'

    def __init__(self, session):
        self.session = session

    def get(self, id, view=None):
        """
        Retrieves a catalog document by ID.

        :param id: the ID of the document to get.
        :param view: the view to get.  One of 'bib', 'client', 'stats', 'all'.
        :return: a :class:`CatalogDocument <mendeley.models.catalog.CatalogDocument>`.
        """
        return super(Catalog, self).get(id, view=view)

    def by_identifier(self, arxiv=None, doi=None, isbn=None, issn=None, pmid=None, scopus=None, filehash=None,
                      view=None):
        """
        Retrieves a catalog document by an external identifier.  Only one identifier may be specified.

        :param arxiv: ArXiV ID.
        :param doi: DOI.
        :param isbn: ISBN.
        :param issn: ISSN.
        :param pmid: PubMed ID.
        :param scopus: Scopus ID (EID).
        :param filehash: SHA-1 filehash.
        :param view: the view to get.  One of 'bib', 'client', 'stats', 'all'.
        :return: a :class:`CatalogDocument <mendeley.models.catalog.CatalogDocument>`.
        """
        url = add_query_params('/catalog', {'arxiv': arxiv, 'doi': doi, 'isbn': isbn, 'issn': issn, 'pmid': pmid,
                                            'scopus': scopus, 'filehash': filehash, 'view': view})
        obj_type = view_type(view)

        rsp = self.session.get(url, headers={'Accept': obj_type.content_type})

        if len(rsp.json()) == 0:
            raise MendeleyException('Catalog document not found')

        return obj_type(self.session, rsp.json()[0])

    def lookup(self, arxiv=None, doi=None, pmid=None, filehash=None, title=None, authors=None, year=None, source=None,
               view=None):
        """
        Finds the closest matching catalog document to a supplied set of metadata.

        :param arxiv: ArXiV ID.
        :param doi: DOI.
        :param pmid: PubMed ID.
        :param filehash: SHA-1 filehash.
        :param title: Title.
        :param authors: Authors.
        :param year: Year.
        :param source: Source.
        :param view: the view to get.  One of 'bib', 'client', 'stats', 'all'.
        :return: a :class:`CatalogDocument <mendeley.models.catalog.CatalogDocument>`.
        """
        url = add_query_params('/metadata', {'arxiv': arxiv, 'doi': doi, 'pmid': pmid, 'filehash': filehash,
                                             'title': title, 'authors': authors, 'year': year, 'source': source})
        obj_type = view_type(view)

        rsp = self.session.get(url, headers={'Accept': obj_type.content_type})

        return LookupResponse(self.session, rsp.json(), view, obj_type)

    def search(self, query, view=None):
        """
        Searches the catalog for documents.

        :param query: the search query to execute.
        :param view: the view to get.  One of 'bib', 'client', 'stats', 'all'.
        :return: a :class:`CatalogSearch <mendeley.models.catalog.CatalogSearch>` resource, from which results can be
                 retrieved.
        """
        return CatalogSearch(self.session, query=query, view=view)

    def advanced_search(self, title=None, author=None, source=None, abstract=None, min_year=None, max_year=None,
                        open_access=None, view=None):
        """
        Executes an advanced search, where individual fields can be searched on.

        :param title: Title.
        :param author: Author.
        :param source: Source.
        :param abstract: Abstract.
        :param min_year: Minimum year for documents to return.
        :param max_year: Maximum year for documents to return.
        :param open_access: If 'true', only returns open access documents.
        :param view: the view to get.  One of 'bib', 'client', 'stats', 'all'.
        :return: a :class:`CatalogSearch <mendeley.models.catalog.CatalogSearch>` resource, from which results can be
                 retrieved.
        """
        return CatalogSearch(self.session, title=title, author=author, source=source, abstract=abstract,
                             min_year=min_year, max_year=max_year, open_access=open_access, view=view)

    @property
    def _session(self):
        return self.session

    def _obj_type(self, **kwargs):
        return view_type(kwargs.get('view'))


class CatalogSearch(ListResource):
    """
    Resource for accessing the results of a catalog search.
    """
    def __init__(self, session, **kwargs):
        self.session = session
        self.params = kwargs

    def list(self, page_size=None):
        """
        Retrieves search results, as a paginated collection.

        :param page_size: the number of search results to return on each page.  Defaults to 20.
        :return: a :class:`Page <mendeley.pagination.Page>` of
                 :class:`CatalogDocuments <mendeley.models.catalog.CatalogDocument>`.
        """
        return super(CatalogSearch, self).list(page_size)

    def iter(self, page_size=None):
        """
        Retrieves search results, as an iterator.

        :param page_size: the number of search results to retrieve at a time.  Defaults to 20.
        :return: an iterator of :class:`CatalogDocuments <mendeley.models.catalog.CatalogDocument>`.
        """
        return super(CatalogSearch, self).iter(page_size)

    def _obj_type(self, **kwargs):
        return view_type(self.params['view'])

    @property
    def _url(self):
        return add_query_params('/search/catalog', self.params)

    @property
    def _session(self):
        return self.session


def view_type(view):
    return {
        'bib': CatalogBibDocument,
        'client': CatalogClientDocument,
        'stats': CatalogStatsDocument,
        'all': CatalogAllDocument,
        'core': CatalogDocument
    }.get(view, CatalogDocument)


import json
from mimetypes import guess_type
from os.path import basename

import arrow

from mendeley.models.base_documents import BaseDocument, BaseBibView, BaseClientView
from mendeley.models.files import File


class UserBaseDocument(BaseDocument):
    @property
    def created(self):
        """
        an :class:`Arrow <arrow.arrow.Arrow>` object.
        """
        if 'created' in self.json:
            return arrow.get(self.json['created'])
        else:
            return None

    @property
    def last_modified(self):
        """
        an :class:`Arrow <arrow.arrow.Arrow>` object.
        """
        if 'last_modified' in self.json:
            return arrow.get(self.json['last_modified'])
        elif 'created' in self.json:
            return arrow.get(self.json['created'])
        else:
            return None

    @property
    def profile(self):
        """
        a :class:`Profile <mendeley.models.profiles.Profile>`.
        """
        if 'profile_id' in self.json:
            return self.session.profiles.get_lazy(self.json['profile_id'])
        else:
            return None

    @property
    def group(self):
        """
        a :class:`Group <mendeley.models.groups.Group>`.
        """
        if 'group_id' in self.json:
            return self.session.groups.get_lazy(self.json['group_id'])
        else:
            return None

    @property
    def files(self):
        """
        a :class:`Files <mendeley.resources.files.Files>` resource, from which
        :class:`Files <mendeley.models.files.File>` can be retrieved.
        """
        return self.session.document_files(document_id=self.id)


class UserBibView(BaseBibView):
    """
    Additional fields returned when getting a :class:`UserDocument <mendeley.models.documents.UserDocument>` or
    :class:`TrashDocument <mendeley.models.documents.TrashDocument>` with view='bib' or 'all'.

    .. attribute:: pages
    .. attribute:: volume
    .. attribute:: issue
    .. attribute:: websites
    .. attribute:: month
    .. attribute:: publisher
    .. attribute:: day
    .. attribute:: city
    .. attribute:: edition
    .. attribute:: institution
    .. attribute:: series
    .. attribute:: chapter
    .. attribute:: revision
    """
    @property
    def accessed(self):
        """
        an :class:`Arrow <arrow.arrow.Arrow>` object.
        """
        if 'accessed' in self.json:
            return arrow.get(self.json['accessed'])
        else:
            return None


class UserClientView(BaseClientView):
    """
    Additional fields returned when getting a :class:`UserDocument <mendeley.models.documents.UserDocument>` or
    :class:`TrashDocument <mendeley.models.documents.TrashDocument>` with view='client' or 'all'.

    .. attribute:: file_attached
    .. attribute:: read
    .. attribute:: starred
    .. attribute:: authored
    .. attribute:: confirmed
    .. attribute:: hidden
    """
    @classmethod
    def fields(cls):
        return super(UserClientView, cls).fields() + ['read', 'starred', 'authored', 'confirmed', 'hidden']


class UserTagsView(object):
    """
    Additional fields returned when getting a :class:`UserDocument <mendeley.models.documents.UserDocument>` or
    :class:`TrashDocument <mendeley.models.documents.TrashDocument>` with view='tags' or 'all'.

    .. attribute:: tags
    """
    @classmethod
    def fields(cls):
        return ['tags']


class UserDocument(UserBaseDocument):
    """
    Base class for user documents.

    .. attribute:: id
    .. attribute:: title
    .. attribute:: type
    .. attribute:: source
    .. attribute:: year
    .. attribute:: identifiers
    .. attribute:: keywords
    .. attribute:: abstract
    """
    def update(self, **kwargs):
        """
        Updates this document.

        :param kwargs: updated field values.  Only the values supplied will be modified.
        :return: the updated document.
        """
        rsp = self.session.patch('/documents/%s' % self.id, data=json.dumps(format_args(kwargs)), headers={
            'Accept': self.content_type,
            'Content-Type': self.content_type
        })

        return UserAllDocument(self.session, rsp.json())

    def delete(self):
        """
        Permanently deletes this document.
        """
        self.session.delete('/documents/%s' % self.id)

    def move_to_trash(self):
        """
        Moves this document to the trash.

        :return: a :class:`TrashDocument <mendeley.models.documents.TrashDocument>`.
        """
        self.session.post('/documents/%s/trash' % self.id)
        return self._trashed_type()(self.session, self.json)

    def attach_file(self, path):
        """
        Attaches a file to this document.

        :param path: the path of the file to attach.
        :return: a :class:`File <mendeley.models.files.File>`.
        """
        filename = basename(path)
        headers = {
            'content-disposition': 'attachment; filename=%s' % filename,
            'content-type': guess_type(filename)[0],
            'link': '<%s/documents/%s>; rel="document"' % (self.session.host, self.id),
            'accept': File.content_type
        }

        with open(path, 'rb') as f:
            rsp = self.session.post('/files', data=f, headers=headers)

        return File(self.session, rsp.json())

    @classmethod
    def _trashed_type(cls):
        return TrashDocument


class UserBibDocument(UserDocument, UserBibView):
    @classmethod
    def fields(cls):
        return UserDocument.fields() + UserBibView.fields()

    @classmethod
    def _trashed_type(cls):
        return TrashBibDocument


class UserClientDocument(UserDocument, UserClientView):
    @classmethod
    def fields(cls):
        return UserDocument.fields() + UserClientView.fields()

    @classmethod
    def _trashed_type(cls):
        return TrashClientDocument


class UserTagsDocument(UserDocument, UserTagsView):
    @classmethod
    def fields(cls):
        return UserDocument.fields() + UserTagsView.fields()

    @classmethod
    def _trashed_type(cls):
        return TrashTagsDocument


class UserAllDocument(UserDocument, UserBibView, UserClientView, UserTagsView):
    @classmethod
    def fields(cls):
        return UserDocument.fields() + UserBibView.fields() + UserClientView.fields() + UserTagsView.fields()

    @classmethod
    def _trashed_type(cls):
        return TrashAllDocument


class TrashDocument(UserBaseDocument):
    """
    Base class for trashed documents.

    .. attribute:: id
    .. attribute:: title
    .. attribute:: type
    .. attribute:: source
    .. attribute:: year
    .. attribute:: identifiers
    .. attribute:: keywords
    .. attribute:: abstract
    """

    def delete(self):
        """
        Permanently deletes this document.
        """
        url = '/trash/%s' % self.id
        self.session.delete(url)

    def restore(self):
        """
        Restores this document from the trash.

        :return: a :class:`UserDocument <mendeley.models.documents.UserDocument>`.
        """
        self.session.post('/trash/%s/restore' % self.id)
        return self._restored_type()(self.session, self.json)

    @classmethod
    def _restored_type(cls):
        return UserDocument


class TrashBibDocument(TrashDocument, UserBibView):
    @classmethod
    def fields(cls):
        return TrashDocument.fields() + UserBibView.fields()

    @classmethod
    def _restored_type(cls):
        return UserBibDocument


class TrashClientDocument(TrashDocument, UserClientView):
    @classmethod
    def fields(cls):
        return TrashDocument.fields() + UserClientView.fields()

    @classmethod
    def _restored_type(cls):
        return UserClientDocument


class TrashTagsDocument(TrashDocument, UserTagsView):
    @classmethod
    def fields(cls):
        return TrashDocument.fields() + UserTagsView.fields()

    @classmethod
    def _restored_type(cls):
        return UserTagsDocument


class TrashAllDocument(TrashDocument, UserBibView, UserClientView, UserTagsView):
    @classmethod
    def fields(cls):
        return TrashDocument.fields() + UserBibView.fields() + UserClientView.fields() + UserTagsView.fields()

    @classmethod
    def _restored_type(cls):
        return UserAllDocument


def format_args(kwargs):
    if 'authors' in kwargs:
        kwargs['authors'] = [author.json for author in kwargs['authors']]
    if 'editors' in kwargs:
        kwargs['editors'] = [editor.json for editor in kwargs['editors']]
    if 'accessed' in kwargs:
        kwargs['accessed'] = arrow.get(kwargs['accessed']).format('YYYY-MM-DD')

    return kwargs


from mendeley.models.documents import *
from mendeley.resources.base_documents import DocumentsBase


class Documents(DocumentsBase):
    """
    Top-level resource for accessing documents.  These can be:

    - documents for the logged-in user, if retrieved from a
      :func:`MendeleySession <mendeley.session.MendeleySession.documents>`.
    - documents in a :func:`Group <mendeley.models.groups.Group.documents>`.
    """
    _url = '/documents'

    def __init__(self, session, group_id):
        super(Documents, self).__init__(session, group_id)

    def get(self, id, view=None):
        """
        Retrieves a document by ID.

        :param id: the ID of the document to get.
        :param view: the view to get.  One of 'bib', 'client', 'tags', 'all'.
        :return: a :class:`UserDocument <mendeley.models.documents.UserDocument>`.
        """
        return super(Documents, self).get(id, view)

    def list(self, page_size=None, view=None, sort=None, order=None, modified_since=None, deleted_since=None):
        """
        Retrieves documents, as a paginated collection.

        :param page_size: the number of documents to return on each page.  Defaults to 20.
        :param view: the view to get.  One of 'bib', 'client', 'tags', 'all'.
        :param sort: if specified, sorts documents by the specified field.  One of 'created', 'last_modified', 'title'.
        :param order: if specified in conjunction with 'sort', specifies the sort order.  One of 'asc', 'desc'.
        :param modified_since: if specified, only returns files modified after this timestamp.
        :param deleted_since: if specified, only returns the IDs of documents deleted after this timestamp.
        :return: a :class:`Page <mendeley.pagination.Page>` of
                 :class:`UserDocuments <mendeley.models.documents.UserDocument>`.
        """
        return super(Documents, self).list(page_size, view, sort, order, modified_since, deleted_since)

    def iter(self, page_size=None, view=None, sort=None, order=None, modified_since=None, deleted_since=None):
        """
        Retrieves documents, as an iterator.

        :param page_size: the number of documents to retrieve at a time.  Defaults to 20.
        :param view: the view to get.  One of 'bib', 'client', 'tags', 'all'.
        :param sort: if specified, sorts documents by the specified field.  One of 'created', 'last_modified', 'title'.
        :param order: if specified in conjunction with 'sort', specifies the sort order.  One of 'asc', 'desc'.
        :param modified_since: if specified, only returns files modified after this timestamp.
        :param deleted_since: if specified, only returns the IDs of documents deleted after this timestamp.
        :return: an iterator of :class:`UserDocuments <mendeley.models.documents.UserDocument>`.
        """
        return super(Documents, self).iter(page_size, view, sort, order, modified_since, deleted_since)

    def create(self, title, type, **kwargs):
        """
        Creates a new document from metadata.

        :param title: title of the document.
        :param type: type of the document.
        :param kwargs: other properties of the document.  These can be any of the attributes on
               :class:`UserDocument <mendeley.models.documents.UserDocument>` or any of its views.
        :return: a :class:`UserDocument <mendeley.models.documents.UserDocument>`.
        """
        kwargs['title'] = title
        kwargs['type'] = type
        kwargs['group_id'] = self.group_id

        kwargs = format_args(kwargs)
        content_type = UserDocument.content_type

        rsp = self.session.post(self._url, data=json.dumps(kwargs), headers={
            'Accept': content_type,
            'Content-Type': content_type
        })

        return UserAllDocument(self.session, rsp.json())

    def create_from_file(self, path):
        """
        Creates a new document from a file.

        :param path: path to the file.
        :return: a :class:`UserDocument <mendeley.models.documents.UserDocument>`.
        """
        filename = basename(path)
        headers = {
            'content-disposition': 'attachment; filename=%s' % filename,
            'content-type': guess_type(filename)[0],
            'accept': UserDocument.content_type
        }

        with open(path, 'rb') as f:
            rsp = self.session.post('/documents', data=f, headers=headers)

        return UserAllDocument(self.session, rsp.json())

    @staticmethod
    def _view_type(view):
        return {
            'all': UserAllDocument,
            'bib': UserBibDocument,
            'client': UserClientDocument,
            'tags': UserTagsDocument,
            'core': UserDocument,
        }.get(view, UserDocument)
